/* Copyright 1991 Digital Equipment Corporation.
 * All Rights Reserved.
 *
 * History:
 *  SCG  21  Tue Jun  2 14:15:36 1992
 *    added newTrace which allows a trace line to be one function call
 *  SCG  14  Wed May 27 13:37:51 1992
 *    added reportAndAbort() which is like report_error followed by
 *    an c_abort.
 *****************************************************************/
/* 	$Id: error.c,v 1.3 1995/07/11 01:53:03 duchier Exp $	 */

#ifndef lint
static char vcid[] = "$Id: error.c,v 1.3 1995/07/11 01:53:03 duchier Exp $";
#endif /* lint */

#include "defs.h"



/* Depth of goal stack */
static long depth_gs()
{
  long i=0;
  ptr_goal g=goal_stack;

  while (g) { i++; g=g->next; }
  return i;
}


/* Depth of choice point stack */
static long depth_cs()
{
  long i=0;
  ptr_choice_point c=choice_stack;

  while (c) { i++; c=c->next; }
  return i;
}


/* Depth of trail (undo) stack */
static long depth_ts()
{
  ptr_stack t=undo_stack;
  long i=0;

  while (t) { i++; t=t->next; }
  return i;
}


void stack_info(outfile)
     FILE *outfile;
{
  /* Information about size of embedded stacks */
  if (verbose) {
    long gn,cn,tn;
    fprintf(outfile,"*** Stack depths [");
    gn=depth_gs();
    cn=depth_cs();
    tn=depth_ts();
    fprintf(outfile,"%ld goal%s, %ld choice point%s, %ld trail entr%s",
            gn,(gn!=1?"s":""),
            cn,(cn!=1?"s":""),
            tn,(tn!=1?"ies":"y"));
    fprintf(outfile,"]\n");
  }
}




void outputline(char *format,...)
{
  va_list VarArg;
  int l;	    
  char buffer[5];   
  char *p; 
  unsigned long lng2; 
  char *cptr; 
  ptr_int_list pil;  
  ptr_psi_term psi;  
  operator kind; 
  def_type t ; 
  va_start(VarArg,format);
  //  vinfoline(format,output_stream, VarArg);
  //  #define vinfoline(format, outfile, xxxx)  {	
  for (p=format;p &&  *p; p++) 
    { 
      if (*p == '%') 
	{ 
	  p++; 
	  switch (*p) 
	    { 
	    case 'd': 
	    case 'x': 
	      buffer[0] = '%'; 
	      buffer[1] = *p; 
	      buffer[2] = 0; 
	      lng2 = va_arg(VarArg,unsigned long); 
	      fprintf(output_stream, buffer, lng2); 
	      break; 
	    case 's': 
	      buffer[0] = '%'; 
	      buffer[1] = *p; 
	      buffer[2] = 0; 
	      cptr = va_arg(VarArg,char *); 
	      fprintf(output_stream, buffer, cptr); 
	      break; 
	    case 'C': 
	      /* type coding as bin string */ 
	      pil = va_arg(VarArg,ptr_int_list); 
	      print_code(output_stream,pil); 
	      break; 
	    case 'P': 
	      psi = va_arg(VarArg,ptr_psi_term); 
	      display_psi(output_stream,psi); 
	      break; 
	    case 'O': 
	      kind = va_arg(VarArg,operator); 
	      print_operator_kind(output_stream,kind); 
	      break; 
	    case 'T': 
	      assert(output_stream==stderr); 
	      t = va_arg(VarArg,def_type); 
	      print_def_type(t); 
	      break; 
	    case 'E': 
	      assert(output_stream==stderr); 
	      perr_i("near line %ld",psi_term_line_number);	
	      if (lf_strcmp(input_file_name,"stdin")) {	   
		perr_s(" in file \042%s\042",input_file_name); 
	      }						   
	      parse_ok=FALSE;					   
	      break;						   
	    case '%':						   
	      putc(*p,output_stream);				   
	      break;						   
	    default: 
	      fprintf(output_stream,"<%c follows %% : report bug >", *p); 
	      break; 
	    } 
	} 
      else 
	putc(*p,output_stream); 
    } 
  va_end(VarArg); 
} 

void Traceline(char *format,...)
{
  va_list VarArg;
  int l;	    
  char buffer[5];   
  char *p; 
  unsigned long lng2; 
  char *cptr; 
  ptr_int_list pil;  
  ptr_psi_term psi;  
  operator kind; 
  def_type t ; 

  va_start(VarArg,format);
  
  if (trace) // DJD
    {
      /* RM: Nov 10 1993  */
      
      if ((trace == 2) && (format[0] != 'p')) return;
      tracing();
      
      //      vinfoline(format, stdout, VarArg);
      // #define vinfoline(format, outfile, xxxx)  { 
      for (p=format;p &&  *p; p++) 
	{ 
	  if (*p == '%') 
	    { 
	      p++; 
	      switch (*p) 
		{ 
		case 'd': 
		case 'x': 
		  buffer[0] = '%'; 
		  buffer[1] = *p; 
		  buffer[2] = 0; 
		  lng2 = va_arg(VarArg,unsigned long); 
		  fprintf(stdout, buffer, lng2); 
		  break; 
		case 's': 
		  buffer[0] = '%'; 
		  buffer[1] = *p; 
		  buffer[2] = 0; 
		  cptr = va_arg(VarArg,char *); 
		  fprintf(stdout, buffer, cptr); 
		  break; 
		case 'C': 
		  /* type coding as bin string */ 
		  pil = va_arg(VarArg,ptr_int_list); 
		  print_code(stdout,pil); 
		  break; 
		case 'P': 
		  psi = va_arg(VarArg,ptr_psi_term); 
		  display_psi(stdout,psi); 
		  break; 
		case 'O': 
		  kind = va_arg(VarArg,operator); 
		  print_operator_kind(stdout,kind); 
		  break; 
		case 'T': 
		  assert(stdout==stderr); 
		  t = va_arg(VarArg,def_type); 
		  print_def_type(t); 
		  break; 
		case 'E': 
		  assert(stdout==stderr); 
		  perr_i("near line %ld",psi_term_line_number);	
		  if (lf_strcmp(input_file_name,"stdin")) {	   
		    perr_s(" in file \042%s\042",input_file_name); 
		  }						   
		  parse_ok=FALSE;					   
		  break;						   
		case '%':						   
		  putc(*p,stdout);				   
		  break;						   
		default: 
		  fprintf(stdout,"<%c follows %% : report bug >", *p); 
		  break; 
		} 
	    } 
	  else 
	    putc(*p,stdout); 
	}
    } 
  va_end(VarArg); 
} 

void Infoline(char *format,...)
{
  va_list VarArg;
  int l;	    
  char buffer[5];   
  char *p; 
  unsigned long lng2; 
  char *cptr; 
  ptr_int_list pil;  
  ptr_psi_term psi;  
  operator kind; 
  def_type t ; 
  
  va_start(VarArg,format);
  if (NOTQUIET)
    {
      
      //      vinfoline(format, stdout, VarArg);
      //#define vinfoline(format, outfile, xxxx)  {	
      for (p=format;p &&  *p; p++) 
	{ 
	  if (*p == '%') 
	    { 
	      p++; 
	      switch (*p) 
		{ 
		case 'd': 
		case 'x': 
		  buffer[0] = '%'; 
		  buffer[1] = *p; 
		  buffer[2] = 0; 
		  lng2 = va_arg(VarArg,unsigned long); 
		  fprintf(stdout, buffer, lng2); 
		  break; 
		case 's': 
		  buffer[0] = '%'; 
		  buffer[1] = *p; 
		  buffer[2] = 0; 
		  cptr = va_arg(VarArg,char *); 
		  fprintf(stdout, buffer, cptr); 
		  break; 
		case 'C': 
		  /* type coding as bin string */ 
		  pil = va_arg(VarArg,ptr_int_list); 
		  print_code(stdout,pil); 
		  break; 
		case 'P': 
		  psi = va_arg(VarArg,ptr_psi_term); 
		  display_psi(stdout,psi); 
		  break; 
		case 'O': 
		  kind = va_arg(VarArg,operator); 
		  print_operator_kind(stdout,kind); 
		  break; 
		case 'T': 
		  assert(stdout==stderr); 
		  t = va_arg(VarArg,def_type); 
		  print_def_type(t); 
		  break; 
		case 'E': 
		  assert(stdout==stderr); 
		  perr_i("near line %ld",psi_term_line_number);	
		  if (lf_strcmp(input_file_name,"stdin")) {	   
		    perr_s(" in file 042%s042",input_file_name); 
		  }						   
		  parse_ok=FALSE;					   
		  break;						   
		case '%':						   
		  putc(*p,stdout);				   
		  break;						   
		default: 
		  fprintf(stdout,"<%c follows %% : report bug >", *p); 
		  break; 
		} 
	    } 
	  else 
	    putc(*p,stdout); 
	}
    } 
  va_end(VarArg); 
} 
void warningline(char *format,...)
{
  va_list VarArg;
  int l;		      
  char buffer[5];	      
  char *p;	      
  unsigned long lng2;   
  char *cptr;	      
  ptr_int_list pil;     
  ptr_psi_term psi;     
  operator kind;	      
  def_type t ;	      
  
  va_start(VarArg,format);
  
  if (warningflag) { // macro would not work
    
    
    if(quietflag) return; /*  RM: Sep 24 1993  */
    fprintf(stderr,"*** Warning: ");
    //    vinfoline(format, stderr, VarArg);
    // #define vinfoline(format, stderr, xxxx)  {	
    for (p=format;p &&  *p; p++) 
      { 
	if (*p == '%') 
	  { 
	    p++; 
	    switch (*p) 
	      { 
	      case 'd': 
	      case 'x': 
		buffer[0] = '%'; 
		buffer[1] = *p; 
		buffer[2] = 0; 
		lng2 = va_arg(VarArg,unsigned long); 
		fprintf(stderr, buffer, lng2); 
		break; 
	      case 's': 
		buffer[0] = '%'; 
		buffer[1] = *p; 
		buffer[2] = 0; 
		cptr = va_arg(VarArg,char *); 
		fprintf(stderr, buffer, cptr); 
		break; 
	      case 'C': 
		/* type coding as bin string */ 
		pil = va_arg(VarArg,ptr_int_list); 
		print_code(stderr,pil); 
		break; 
	      case 'P': 
		psi = va_arg(VarArg,ptr_psi_term); 
		display_psi(stderr,psi); 
		break; 
	      case 'O': 
		kind = va_arg(VarArg,operator); 
		print_operator_kind(stderr,kind); 
		break; 
	      case 'T': 
		assert(stderr==stderr); 
		t = va_arg(VarArg,def_type); 
		print_def_type(t); 
		break; 
	      case 'E': 
		assert(stderr==stderr); 
		perr_i("near line %ld",psi_term_line_number);	
		if (lf_strcmp(input_file_name,"stdin")) {	   
		  perr_s(" in file 042%s042",input_file_name); 
		}						   
		parse_ok=FALSE;					   
		break;						   
	      case '%':						   
		putc(*p,stderr);				   
		break;						   
	      default: 
		fprintf(stderr,"<%c follows %% : report bug >", *p); 
		break; 
	      } 
	  } 
	else 
	  putc(*p,stderr); 
      } 
    va_end(VarArg); 
  } 
}

/* New error printing routine */
void Errorline(char *format,...)
{
  va_list VarArg;
  int l;	    
  char buffer[5];   
  char *p; 
  unsigned long lng2; 
  char *cptr; 
  ptr_int_list pil;  
  ptr_psi_term psi;  
  operator kind; 
  def_type t ; 
  
  va_start(VarArg,format);
  //  fprintf(stderr,"format = %lx %s\n",(long)format,format);fflush(stdout);
  fprintf(stderr,"*** Error: ");
  //  fprintf(stderr,"format2 = %lx %s\n",(long)format,format);
  // vinfoline(format, stderr, VarArg);
  //#define vinfoline(format, stderr, xxxx)  { 
  for (p=format;p &&  *p; p++) 
    { 
      if (*p == '%') 
	{ 
	  p++; 
	  switch (*p) 
	    { 
	    case 'd': 
	    case 'x': 
	      buffer[0] = '%'; 
	      buffer[1] = *p; 
	      buffer[2] = 0; 
	      lng2 = va_arg(VarArg,unsigned long); 
	      fprintf(stderr, buffer, lng2); 
	      break; 
	    case 's': 
	      buffer[0] = '%'; 
	      buffer[1] = *p; 
	      buffer[2] = 0; 
	      cptr = va_arg(VarArg,char *); 
	      fprintf(stderr, buffer, cptr); 
	      break; 
	    case 'C': 
	      /* type coding as bin string */ 
	      pil = va_arg(VarArg,ptr_int_list); 
	      print_code(stderr,pil); 
	      break; 
	    case 'P': 
	      psi = va_arg(VarArg,ptr_psi_term); 
	      display_psi(stderr,psi); 
	      break; 
	    case 'O': 
	      kind = va_arg(VarArg,operator); 
	      print_operator_kind(stderr,kind); 
	      break; 
	    case 'T': 
	      assert(stderr==stderr); 
	      t = va_arg(VarArg,def_type); 
	      print_def_type(t); 
	      break; 
	    case 'E': 
	      assert(stderr==stderr); 
	      perr_i("near line %ld",psi_term_line_number);	
	      if (lf_strcmp(input_file_name,"stdin")) {	   
		perr_s(" in file \042%s\042",input_file_name); 
	      }						   
	      parse_ok=FALSE;					   
	      break;						   
	    case '%':						   
	      putc(*p,stderr);				   
	      break;						   
	    default: 
	      fprintf(stderr,"<%c follows %% : report bug >", *p); 
	      break; 
	    } 
	} 
      else 
	putc(*p,stderr); 
    } 
  va_end(VarArg); 
#ifdef CLIFE
exit(0);
#endif
}
void Syntaxerrorline(char *format,...)
{
  va_list VarArg;
  int l;	    
  char buffer[5];   
  char *p; 
  unsigned long lng2; 
  char *cptr; 
  ptr_int_list pil;  
  ptr_psi_term psi;  
  operator kind; 
  def_type t ; 
  va_start(VarArg,format);
  //  fprintf(stderr,"format = %lx %s\n",(long)format,format);fflush(stdout);
  if(parse_ok) { /*  RM: Feb  1 1993  */
    parse_ok=FALSE; /*  RM: Feb  1 1993  */
    fprintf(stderr,"*** Syntax error: ");
    //    fprintf(stderr,"format2 = %lx %s\n",(long)format,format);
    // vinfoline(format, stderr, VarArg);
    //#define vinfoline(format, outfile, xxxx)  {	
    for (p=format;p &&  *p; p++) 
      { 
	if (*p == '%') 
	  { 
	    p++; 
	    switch (*p) 
	      { 
	      case 'd': 
	      case 'x': 
		buffer[0] = '%'; 
		buffer[1] = *p; 
		buffer[2] = 0; 
		lng2 = va_arg(VarArg,unsigned long); 
		fprintf(stderr, buffer, lng2); 
		break; 
	      case 's': 
		buffer[0] = '%'; 
		buffer[1] = *p; 
		buffer[2] = 0; 
		cptr = va_arg(VarArg,char *); 
		fprintf(stderr, buffer, cptr); 
		break; 
	      case 'C': 
		/* type coding as bin string */ 
		pil = va_arg(VarArg,ptr_int_list); 
		print_code(stderr,pil); 
		break; 
	      case 'P': 
		psi = va_arg(VarArg,ptr_psi_term); 
		display_psi(stderr,psi); 
		break; 
	      case 'O': 
		kind = va_arg(VarArg,operator); 
		print_operator_kind(stderr,kind); 
		break; 
	      case 'T': 
		assert(stderr==stderr); 
		t = va_arg(VarArg,def_type); 
		print_def_type(t); 
		break; 
	      case 'E': 
		assert(stderr==stderr); 
		perr_i("near line %ld",psi_term_line_number);	
		if (lf_strcmp(input_file_name,"stdin")) {	   
		  perr_s(" in file \042%s\042",input_file_name); 
		}						   
		parse_ok=FALSE;					   
		break;						   
	      case '%':						   
		putc(*p,stderr);				   
		break;						   
	      default: 
		fprintf(stderr,"<%c follows %% : report bug >", *p); 
		break; 
	      } 
	  } 
	else 
	  putc(*p,stderr); 
      } 
    va_end(VarArg); 
  } 
}

/********************************************************************/

/* Utilities for tracing and single stepping */

/* Initialize all tracing variables */
void init_trace()
{
  trace=FALSE;
  stepflag=FALSE;
  stepcount=0;
}

/* Reset stepcount to zero */
/* Should be called when prompt is printed */
void reset_step()
{
  if (stepcount>0) {
    stepcount=0;
    stepflag=TRUE;
  }
}

void tracing()
{
  long i;
  long indent;

  printf("T%04ld",goal_count);
  printf(" C%02ld",depth_cs());
  indent=depth_gs();
  if (indent>=MAX_TRACE_INDENT) printf(" G%02ld",indent);
  indent = indent % MAX_TRACE_INDENT;
  for (i=indent; i>=0; i--) printf(" ");
  steptrace=TRUE;
}


void new_trace(newtrace)
     long newtrace;
{
  trace = newtrace;
  printf("*** Tracing is turned ");
  printf(trace?"on.":"off.");
  if (trace == 2) printf(" Only for Proves");
  printf("\n");
}

void new_step(newstep)
     long newstep;
{
  stepflag = newstep;
  printf("*** Single stepping is turned ");
  printf(stepflag?"on.\n":"off.\n");
  new_trace(stepflag);
  steptrace=FALSE;
}

void set_trace_to_prove()
{
  new_trace(2);
}

void toggle_trace()
{
  new_trace(trace?0:1);
}


void toggle_step()
{
  new_step(!stepflag);
}

/********************************************************************/

/* Old error printing routines -- these should be superceded by Errorline */

void perr(str)
     char *str;
{
  fputs(str, stderr);
}

void perr_s(s1,s2)
     char *s1,*s2;
{
  fprintf(stderr,s1,s2);
}

void perr_s2(s1,s2,s3)
     char *s1,*s2,*s3;
{
  fprintf(stderr,s1,s2,s3);
}

void perr_i(str,i)
     char *str;
     long i;
{
  fprintf(stderr,str,i);
}


long warning()
{
  if (warningflag) perr("*** Warning: ");
  return warningflag;
}


long warningx()
{
  if (warningflag) perr("*** Warning");
  return warningflag;
}


/* Main routine for report_error and report_warning */
void report_error_main(g,s,s2)
     ptr_psi_term g;
     char *s, *s2;
{
  FILE *f;

  perr_s2("*** %s: %s in '",s2,s);
  display_psi_stderr(g);
  perr("'.\n");
}



/******** REPORT_ERROR(g,s)
	  Print an appropriate error message. G is the
	  psi-term which caused the error, S a message to print.
	  Format: '*** Error: %s in 'g'.'
*/
void report_error(g,s)
     ptr_psi_term g;
     char *s;
{
  report_error_main(g,s,"Error");
}



/******** REPORTANDABORT(g,s)
	  Print an appropriate error message. G is the
	  psi-term which caused the error, S a message to print.
	  Format: '*** Error: %s in 'g'.'
*/
long reportAndAbort(g,s)
     ptr_psi_term g;
     char *s;
{
  report_error_main(g,s,"Error");
  return abort_life(TRUE); // djd added TRUE
}


/******** REPORT_WARNING(g,s)
	  Print an appropriate error message. G is the
	  psi-term which caused the error, S a message to print.
	  Format: '*** Warning: %s in 'g'.'
*/
void report_warning(g,s)
     ptr_psi_term g;
     char *s;
{
  if (warningflag) report_error_main(g,s,"Warning");
}


/* Main routine for report_error2 and report_warning2 */
void report_error2_main(g,s,s2)
     ptr_psi_term g;
     char *s, *s2;
{
  FILE *f;

  perr_s("*** %s: argument '",s2);
  display_psi_stderr(g);
  perr_s("' %s.\n",s);
}



/********* REPORT_ERROR2(g,s)
	   Like report_error, with a slightly different format.
	   Format: '*** Error: argument 'g' %s.'
*/
void report_error2(g,s)
     ptr_psi_term g;
     char *s;
{
  report_error2_main(g,s,"Error");
}



/********* REPORT_WARNING2(g,s)
	   Like report_warning, with a slightly different format.
	   Format: '*** Warning: argument 'g' %s.'
*/
void report_warning2(g,s)
     ptr_psi_term g;
     char *s;
{
  if (warningflag) report_error2_main(g,s,"Warning");
}



/* Give error message if there is an argument which cannot unify with */
/* a real number. */
void nonnum_warning(t,arg1,arg2)
     ptr_psi_term t,arg1,arg2;
{
  if (!curried && /* PVR 15.9.93 */
      ((arg1 && !overlap_type(arg1->type,real)) ||
       (arg2 && !overlap_type(arg2->type,real)))) {
    report_warning(t,"non-numeric argument(s)");
  }
}

/********************************************************************/

/* Error checking routines for bit_and, bit_or, shift, and modulo */

long nonint_warning(arg, val, msg)
     ptr_psi_term arg;
     REAL val;
     char *msg;
{
  long err=FALSE;

  if (val!=floor(val)) {
    report_warning2(arg, msg);
    err=TRUE;
  }
  return err;
}

long bit_and_warning(arg, val)
     ptr_psi_term arg;
     REAL val;
{
  return nonint_warning(arg,val,"of bitwise 'and' operation is not an integer");
}

long bit_or_warning(arg, val)
     ptr_psi_term arg;
     REAL val;
{
  return nonint_warning(arg,val,"of bitwise 'or' operation is not an integer");
}

long bit_not_warning(arg, val)
     ptr_psi_term arg;
     REAL val;
{
  return nonint_warning(arg,val,"of bitwise 'not' operation is not an integer");
}

long int_div_warning(arg, val)
     ptr_psi_term arg;
     REAL val;
{
  return nonint_warning(arg,val,"of integer division is not an integer");
}

long mod_warning(arg, val,zero)
     ptr_psi_term arg;
     REAL val;
     int zero;
{
  int err;

  err=nonint_warning(arg,val,"of modulo operation is not an integer");
  if(!err && zero && val==0) {
    Errorline("division by 0 in modulo operation\n");
    err=TRUE;
  }
  return err;
}

long shift_warning(dir, arg, val)
     long dir;
     ptr_psi_term arg;
     REAL val;
{
  if (dir)
    return nonint_warning(arg,val,"of right shift operation is not an integer");
  else
    return nonint_warning(arg,val,"of left shift operation is not an integer");
}

/********************************************************************/
